from dataclasses import asdict, is_dataclass, dataclass
from enum import Enum
import time
from typing import Awaitable, Callable, Generic, Optional, TypeVar
import multiprocessing
from contextlib import asynccontextmanager
from os import getenv

from starlette.applications import Starlette
from starlette.responses import (
    JSONResponse as StarletteJSONResponse,
    Response as StarletteResponse,
)
from starlette.requests import Request as StarletteRequest
from starlette.routing import Route
from starlette.middleware.base import BaseHTTPMiddleware, RequestResponseEndpoint

import uvicorn

from onbbu.database import DatabaseManager, database

from pydantic import ValidationError

from onbbu.logger import LogLevel, logger

T = TypeVar("T")


class Request(StarletteRequest):
    pass


class JSONResponse(StarletteJSONResponse):
    pass


class ResponseNotFoundError(JSONResponse):

    def render(self, content: str) -> bytes:
        content = {"error": content}  # type: ignore

        return super().render(content, status_code=404)  # type: ignore


class ResponseValueError(JSONResponse):

    def render(self, content: ValueError) -> bytes:
        content = {"error": str(content)}  # type: ignore

        return super().render(content, status_code=500)  # type: ignore


class ResponseValidationError(JSONResponse):

    def render(self, content: ValidationError) -> bytes:
        content = {"detail": content.errors()}  # type: ignore

        return super().render(content, status_code=400)  # type: ignore


class Response(Generic[T], JSONResponse):

    def render(self, content: T) -> bytes:  # type: ignore

        if is_dataclass(content):
            content = asdict(content)  # type: ignore

        elif isinstance(content, list):
            content = [
                (asdict(item) if is_dataclass(item) else item) for item in content  # type: ignore
            ]

        return super().render(content)


class TimingMiddleware(BaseHTTPMiddleware):
    async def dispatch(
        self, request: StarletteRequest, call_next: RequestResponseEndpoint
    ) -> StarletteResponse:
        start_time: float = time.time()

        response: StarletteResponse = await call_next(request)

        process_time: float = time.time() - start_time

        response.headers["X-Process-Time"] = str(process_time)

        return response


class HTTPMethod(Enum):
    GET = "GET"
    POST = "POST"
    PUT = "PUT"
    DELETE = "DELETE"


EndpointHttpType = Callable[[Request], Awaitable[JSONResponse]]


@dataclass(frozen=True, slots=True)
class RouteDTO:
    path: str
    endpoint: EndpointHttpType
    method: HTTPMethod


class RouterHttp:
    __router: list[Route]
    __prefix: str

    def __init__(self, prefix: str = ""):
        self.__router = []
        self.__prefix = prefix.rstrip("/")

    def add_route(self, dto: RouteDTO) -> None:

        full_path: str = f"{self.__prefix}{dto.path}"

        self.__router.append(
            Route(path=full_path, endpoint=dto.endpoint, methods=[dto.method.value])
        )

    def add_routes(self, dtos: list[RouteDTO]) -> None:
        for dto in dtos:
            self.add_route(dto)

    def get_router(self) -> list[Route]:
        return self.__router

    def get_routes(self) -> list[str]:
        return [route.path for route in self.__router]


class ServerHttp:
    database: DatabaseManager
    host: str
    port: int
    environment: str
    reload: bool
    workers: int
    server: Starlette

    def __init__(self, environment: str, port: Optional[int]):
        self.host = "0.0.0.0"
        self.port = port or 8000
        self.environment = environment
        self.reload = self.environment == "development"
        self.workers = 1 if self.reload else max(2, multiprocessing.cpu_count() - 1)

        self.server = Starlette(debug=True, routes=[], lifespan=self._lifespan)

        self.server.add_middleware(TimingMiddleware)

        self.database = database

    @asynccontextmanager
    async def _lifespan(self, app: Starlette):
        """Gestor de eventos de vida para FastAPI"""
        await self.database.init()
        yield
        await self.database.close()

    def include_router(self, router: RouterHttp) -> None:
        """Agrega todas las rutas de un RouterHttp a la aplicación"""
        self.server.router.routes.extend(router.get_router())


server_http: ServerHttp = ServerHttp(
    port=int(getenv("HTTP_PORT", "8000")),
    environment=getenv("ENVIRONMENT", "development"),
)


def runserver(server_http: ServerHttp) -> None:
    logger.log(
        level=LogLevel.INFO,
        message=f"🚀 Iniciando servidor en {server_http.host}:{server_http.port} ...",
        extra_data={},
    )

    for route in server_http.server.routes:
        logger.log(
            level=LogLevel.INFO,
            message=f"🔗 {route.path} -> {route.name} ({route.methods})",  # type: ignore
            extra_data={},
        )

    uvicorn.run(server_http.server, host=server_http.host, port=server_http.port)



from enum import Enum
from typing import Awaitable, Callable

Listener = Callable[..., None] | Callable[..., Awaitable[None]]

class Event(Enum):
    DEFAULT_DISCOUNT = 'add_default_discounts'

class EventEmitter:
    def __init__(self) -> None: ...
    def on(self, event: str, listener: Listener) -> None: ...
    async def emit(self, event: str, *args: object, **kwargs: object) -> None: ...

emitter: EventEmitter



from dataclasses import dataclass
from typing import Generic, TypeVar

T = TypeVar("T")
U = TypeVar("U")

@dataclass(frozen=True, slots=True)
class PaginateDTO:
    page: int
    limit: int

@dataclass(frozen=True, slots=True)
class PaginateOutputDTO(Generic[T]):
    page: int
    limit: int
    total: int
    total_page: int
    data: list[T]

def calculate_total_pages(total: int, limit: int) -> int: ...



from argparse import ArgumentParser, Namespace
from typing import List
from onbbu.database import database as database
from onbbu.http import server_http as server_http
from onbbu.logger import LogLevel as LogLevel, logger as logger

class BaseCommand:
    name: str
    help: str
    def add_arguments(self, parser: ArgumentParser) -> None: ...
    async def handler(self, args: Namespace) -> None: ...

class MigrateCommand(BaseCommand):
    name: str
    help: str
    async def handler(self, args: Namespace) -> None: ...

class CreateModuleCommand(BaseCommand):
    name: str
    help: str
    def add_arguments(self, parser: ArgumentParser) -> None: ...
    async def handler(self, args: Namespace) -> None: ...

async def cli() -> None: ...

async def menu_cli(description: str, commands: List[BaseCommand]) -> None: ...


from dataclasses import dataclass
from enum import Enum
from onbbu.database import DatabaseManager as DatabaseManager, database as database
from onbbu.logger import LogLevel as LogLevel, logger as logger
from pydantic import ValidationError as ValidationError
from starlette.applications import Starlette
from starlette.middleware.base import BaseHTTPMiddleware, RequestResponseEndpoint as RequestResponseEndpoint
from starlette.requests import Request as StarletteRequest
from starlette.responses import JSONResponse as StarletteJSONResponse, Response as StarletteResponse
from starlette.routing import Route
from typing import Awaitable, Callable, Generic, TypeVar

T = TypeVar('T')

class Request(StarletteRequest): ...
class JSONResponse(StarletteJSONResponse): ...

class ResponseNotFoundError(JSONResponse):
    def render(self, content: str) -> bytes: ...

class ResponseValueError(JSONResponse):
    def render(self, content: ValueError) -> bytes: ...

class ResponseValidationError(JSONResponse):
    def render(self, content: ValidationError) -> bytes: ...

class Response(JSONResponse, Generic[T]):
    def render(self, content: T) -> bytes: ...

class TimingMiddleware(BaseHTTPMiddleware):
    async def dispatch(self, request: StarletteRequest, call_next: RequestResponseEndpoint) -> StarletteResponse: ...

class HTTPMethod(Enum):
    GET = 'GET'
    POST = 'POST'
    PUT = 'PUT'
    DELETE = 'DELETE'
EndpointHttpType = Callable[[Request], Awaitable[JSONResponse]]

@dataclass(frozen=True, slots=True)
class RouteDTO:
    path: str
    endpoint: EndpointHttpType
    method: HTTPMethod

class RouterHttp:
    def __init__(self, prefix: str = '') -> None: ...
    def add_route(self, dto: RouteDTO) -> None: ...
    def add_routes(self, dtos: list[RouteDTO]) -> None: ...
    def get_router(self) -> list[Route]: ...
    def get_routes(self) -> list[str]: ...

class ServerHttp:
    database: DatabaseManager
    host: str
    port: int
    environment: str
    reload: bool
    workers: int
    server: Starlette
    def __init__(self, environment: str, port: int | None) -> None: ...
    def include_router(self, router: RouterHttp) -> None: ...

server_http: ServerHttp

def runserver(server_http: ServerHttp) -> None: ...



from typing import Any, List, Type
from os import getenv
from tortoise import Model, Tortoise
from onbbu.logger import LogLevel, logger
from aerich import Command  # type: ignore


class DatabaseManager:
    database_url: str
    command: Command  # type: ignore

    def __init__(self, database_url: str):
        self.database_url = database_url
        self.models: List[str] = []

    def register_models(self, model: Type[Model]) -> None:
        """Register a model in the database"""
        if model not in self.models:
            self.models.append(model.__module__)

    def get_config(self) -> dict[str, Any]:

        tortoise_config: dict[str, Any] = {
            "connections": {
                "default": self.database_url,
            },
            "apps": {
                "models": {
                    "models": self.models,
                    "default_connection": "default",
                },
            },
        }

        return tortoise_config

    async def init(self) -> None:
        """Initialize the database and apply the migrations."""

        if not self.models:
            logger.log(
                level=LogLevel.ERROR,
                message="❌ No models found. Check Check.",
                extra_data={},
            )

            return

        logger.log(
            level=LogLevel.INFO,
            message=f"🔄 Initializing Tortoise with models: {self.models}",
            extra_data={},
        )

        await Tortoise.init(config=self.get_config())

        self.command = Command(tortoise_config=self.get_config())

        logger.log(
            level=LogLevel.INFO,
            message="✅ Connected database. Generating schematics...",
            extra_data={},
        )

        await Tortoise.generate_schemas(safe=True)

        logger.log(
            level=LogLevel.INFO,
            message="🎉 Schemes generated successfully.",
            extra_data={},
        )

        await Tortoise.close_connections()

    async def migrate(self) -> None:
        """Generate new migrations."""

        await self.command.init()

        await self.command.migrate()

    async def upgrade(self) -> None:
        """Apply pending migrations."""

        await self.command.init()

        await self.command.upgrade()

    async def downgrade(self, steps: int = 1) -> None:
        """Revert migrations (default: 1 step)."""

        await self.command.init()

        await self.command.downgrade(steps, delete=False)

    async def history(self) -> None:
        """Show the history of applied migrations."""

        await self.command.init()

        await self.command.history()

    async def create_database(self) -> None:
        """Create the database if it does not exist."""

        logger.log(
            level=LogLevel.INFO,
            message=f"🛠️ Creating database...",
            extra_data={},
        )

        await Tortoise.init(config=self.get_config())

        await Tortoise.generate_schemas()

        await Tortoise.close_connections()

    async def drop_database(self) -> None:
        """Delete all tables from the database."""

        logger.log(
            level=LogLevel.INFO,
            message="🗑️ Dropping database...",
            extra_data={},
        )

        await Tortoise.init(config=self.get_config())

        await Tortoise.generate_schemas(safe=False)

        await Tortoise.close_connections()

    async def reset_database(self) -> None:
        """Delete and recreate the database."""

        logger.log(
            level=LogLevel.INFO,
            message=f"🔄 Resetting database...",
            extra_data={},
        )

        await self.drop_database()

        await self.create_database()

    async def show_status(self) -> None:
        """Show the current status of the database."""

        await self.command.init()

        applied = await self.command.history()

        logger.log(
            level=LogLevel.INFO,
            message=f"📜 Applied migrations:\n{applied}",
            extra_data={},
        )

    async def apply_all_migrations(self) -> None:
        """Generate and apply all migrations in a single step."""

        logger.log(
            level=LogLevel.INFO,
            message=f"🚀 Applying all migrations...",
            extra_data={},
        )

        await self.migrate()

        await self.upgrade()

    async def rollback_all_migrations(self) -> None:
        """Revert all migrations to the initial state."""

        logger.log(
            level=LogLevel.INFO,
            message=f"⏪ Rolling back all migrations...",
            extra_data={},
        )

        await self.command.init()

        while True:
            try:

                await self.command.downgrade(1, delete=False)

            except Exception:

                logger.log(
                    level=LogLevel.INFO,
                    message=f"✅ No more migrations to revert.",
                    extra_data={},
                )

                break

    async def seed_data(self) -> None:
        """Insert initial data into the database."""

        logger.log(
            level=LogLevel.INFO,
            message=f"🌱 Seeding initial data...",
            extra_data={},
        )

        await Tortoise.init(config=self.get_config())

        await Tortoise.close_connections()

    async def close(self) -> None:
        """Close the database connections."""

        logger.log(
            level=LogLevel.INFO,
            message=f"🔌 Closing database connections...",
            extra_data={},
        )

        await Tortoise.close_connections()

        logger.log(
            level=LogLevel.INFO,
            message=f"✅ Database connections closed..",
            extra_data={},
        )


database: DatabaseManager = DatabaseManager(
    database_url=getenv("DATABASE_URL", "sqlite://db.sqlite3")
)
